var FieldStates = {
    fs_Empty: 0,
    fs_Block: 1
};

var GameStates = {
    gs_Stop: 0,
    gs_Play: 1,
    gs_Loading: 2
};

var Game = {

    /* Nastavenia */

    //Zakladne pocty bodov pre zaplnenie 1-4 riadku
    Line1Score: 40,
    Line2Score: 100,
    Line3Score: 300,
    Line4Score: 1200,

    //Rozmery hracieho pola (pocet policok, nie px)
    GameArrayWidth: 10,
    GameArrayHeight: 20,

    GameFieldWidth: 25, //Velkost jedneho policka
    BorderWidth: 2, //Hrubka oramovania hracieho pola

    BackColor: '#000000', //Farba pozadia
    BorderColor: '#303030', //Farba oramovania

    //Rozmery hracej plochy (v px)
    GameAreaWidth: 263,
    GameAreaHeight: 523,

    //Pozicia obrazku GameOver
    GameOverLeft: 1,
    GameOverTop: 157,

    //Pozicia score panelu
    ScorePanelTop: 280,
    SocrePanelLeft: 5,

    DigitWidth: 12, //Sirka cislice        
    DigitHeight: 22, //Dlzka cislice
    DigitPadding: 2, //Padding medzi cislicami

    //Pozicia na umiestnenie ukazovatela score
    ScoreTop: 5,
    ScoreLeft: 375,

    //Pozicia na umiestnenie ukazovatela levelu
    LevelTop: 43,
    LevelLeft: 375,

    //Pozicia na umiestnenie ukazovatela nasledujucej kocky
    PreviewTop: 80,
    PreviewLeft: 375,

    /* /Nastavenia */

    /* Herne premenne */
    Canvas: null,
    Context: null, //Kresliaca plocha
    ActivePiece: null, //Aktivna kocka
    NextPiece: null, //Nasledujuca kocka
    LastUpdate: null, //Cas poslednej aktualizacie hry
    LastRandomNumber: -1, //Posledne vratene nahodne cislo

    //Rozmery kresliacej plochy
    ContextWidth: 0,
    ContextHeight: 0,

    Score: 0, //Score
    Level: 1, //Level            
    LinesCounter: 0, //Pocet riadkov plnych riadkov
    Status: GameStates.gs_Loading, //Status hry
    GameArray: null, //Herne pole
    ItemsToLoad: 0, //Pocet poloziek, ktore sa este nacitavaju
    /* /Herne premenne */

    /* Obrazky */
    img_empty: null, //Prazde policko
    img_colors: [], //Farebne kocky
    img_gameover: null, //Napis game over
    img_scorepanel: null, //Score panel
    img_digits: [], //Digitalne cislice
    /* /Obrazky */

    /* Zvuky */
    snd_lines: null, //Vyplnenie riadka
    /* /Zvuky */

    LoadGame: function ()
    {
        //Event listener na stlacene klavesy
        window.addEventListener('keydown', this.KeyDown);

        //Nacitanie obrazkov
        this.LoadImages();

        //Nacitanie zvukov
        this.LoadSounds();

        //Inicializacia kresliaceho prostredia
        this.LoadCanvas();

        //Skontrolujeme ci uz nebolo vsetko nahrate
        this.CheckLoadedFiles();
    },
    KeyDown: function (e)
    {
        switch (e.keyCode)
        {
            case 13: //Enter - Restart game
                {
                    if (Game.Status === GameStates.gs_Stop)
                    {
                        Game.NewGame();
                    }

                    e.preventDefault();
                }
                break;
            case 38: //Up - Rotate
                {
                    Game.RotatePiece(Game.ActivePiece);
                    e.preventDefault();
                }
                break;
            case 40: //Down
                {
                    if (Game.ActivePiece.Top + Game.ActivePiece.Height + 1 <= Game.GameArrayHeight)
                    {
                        Game.MovePiece(1, 0, false);
                    }

                    e.preventDefault();
                }
                break;
            case 37: //Left
                {
                    if (Game.ActivePiece.Left - 1 >= 0)
                    {
                        Game.MovePiece(0, -1, false);
                    }

                    e.preventDefault();
                }
                break;
            case 39: //Right
                {
                    if (Game.ActivePiece.Left + Game.ActivePiece.Width + 1 <= Game.GameArrayWidth)
                    {
                        Game.MovePiece(0, 1, false);
                    }

                    e.preventDefault();
                }
                break;
            case 32: //SpaceBar
                {
                    var LowerTop = Game.GetLowerPosition(Game.ActivePiece);
                    if (LowerTop <= 0)
                        return;

                    Game.MovePiece(LowerTop - Game.ActivePiece.Top, 0, false);

                    e.preventDefault();
                }
                break;
        }
    },

    CheckLoadedFiles: function ()
    {
        if (Game.ItemsToLoad <= 0)
        {
            this.NewGame();
        }
    },
    NewGame: function ()
    {
        this.LastUpdate = new Date();

        //Reset hernych premennych
        this.Level = 1;
        this.Score = 0;
        this.LinesCounter = 0;
        this.NextPiece = null;

        //Vycistit hracie pole
        this.CreateGameArray();

        //Nakreslit hracie pole
        this.DrawGameField();

        //Nova kocka na plochu
        this.NewPiece();

        this.Status = GameStates.gs_Play;

        this.GameUpdateRequest();

    },
    CreateImage: function (FileName)
    {
        var img = new Image();
        this.ItemsToLoad++;
        img.src = "imgs/" + FileName + ".png";
        img.onload = this.ImageLoaded;
        return img;
    },
    ImageLoaded: function ()
    {
        this.onload = null;
        Game.ItemsToLoad--;
        Game.CheckLoadedFiles();
    },
    CreateSound: function (Source)
    {
        var Sound = document.createElement("audio");
        Sound.src = "audio/" + Source + '.mp3';
        Sound.setAttribute("preload", "auto");
        Sound.setAttribute("controls", "none");
        this.ItemsToLoad++;
        Sound.oncanplay = this.SoundLoaded;
        return Sound;
    },
    SoundLoaded: function ()
    {
        this.oncanplay = null;
        Game.ItemsToLoad--;
        Game.CheckLoadedFiles();
    },
    LoadImages: function ()
    {
        this.img_empty = this.CreateImage("empty");
        this.img_gameover = this.CreateImage("gameover");

        //Nacitanie farebnych kociek
        this.img_colors[0] = this.CreateImage("blue");
        this.img_colors[1] = this.CreateImage("brown");
        this.img_colors[2] = this.CreateImage("green");
        this.img_colors[3] = this.CreateImage("purple");
        this.img_colors[4] = this.CreateImage("red");
        this.img_colors[5] = this.CreateImage("yellow");

        this.img_scorepanel = this.CreateImage("scorepanel");
        //Nacitanie obrazkov cisel
        for (var i = 0; i <= 9; i++)
            this.img_digits[i] = this.CreateImage("d" + i.toString());
    },
    LoadSounds: function ()
    {
        this.snd_lines = this.CreateSound("lines");
    },
    LoadCanvas: function ()
    {
        this.Canvas = document.getElementById('gamecanvas');
        this.Context = this.Canvas.getContext("2d");

        // Odpamatame si rozmery kresliacej plochy
        this.ContextWidth = this.Canvas.offsetWidth;
        this.ContextHeight = this.Canvas.offsetHeight;
    },
    CreateGameArray: function ()
    {
        var i, j;

        this.GameArray = [];
        for (i = 0; i < this.GameArrayHeight; i++)
        {
            this.GameArray[i] = this.CreateEmptyLine();
        }
    },
    CreateEmptyLine: function ()
    {

        var ret = [];
        for (var j = 0; j < this.GameArrayWidth; j++)
        {
            ret[j] = FieldStates.fs_Empty;
        }

        return ret;
    },
    DrawGameField: function ()
    {
        //Vyplnenie celej plochy farbou pozadia
        this.Context.fillStyle = this.BackColor;
        this.Context.fillRect(0, 0, this.ContextWidth, this.ContextHeight);

        //Vyplnenie hernej plochy farbou oramovania
        this.Context.fillStyle = this.BorderColor;
        this.Context.fillRect(0, 0, this.GameAreaWidth, this.GameAreaHeight);

        //Nakreslenie jednotlivych policok
        for (var i = 0; i < this.GameArrayHeight; i++)
        {
            for (var j = 0; j < this.GameArrayWidth; j++)
            {
                this.DrawBlock(i, j, this.img_empty);
            }
        }

        //Nakreslenie score panelu
        this.Context.drawImage(this.img_scorepanel, this.ScorePanelTop, this.SocrePanelLeft);

        //Aktualizacia score hodnoty
        this.UpdateScore();
    },

    GetRealPosition: function (pos)
    {
        return this.BorderWidth + pos * this.GameFieldWidth + pos;
    },
    DrawBlock: function (top, left, color)
    {
        var RealLeft = this.GetRealPosition(left);
        var RealTop = this.GetRealPosition(top);
        this.Context.drawImage(color, RealLeft, RealTop);
    },
    DrawBlockInPreview: function (top, left, color)
    {
        var RealLeft = this.PreviewLeft + left * this.GameFieldWidth + left;
        var RealTop = this.PreviewTop + top * this.GameFieldWidth + top;
        this.Context.drawImage(color, RealLeft, RealTop);
    },
    CreatePiece: function (width, height, color)
    {
        var Ret = {Left: 0, Top: 0, Width: width, Height: height, Color: color, Blocks: []};

        Ret.Blocks = [];
        for (var i = 0; i < height; i++)
        {
            Ret.Blocks[i] = [];
            for (var j = 0; j < width; j++)
            {
                Ret.Blocks[i][j] = FieldStates.fs_Empty;
            }
        }

        return Ret;
    },
    Create_L_Piece: function (color)
    {
        var piece = this.CreatePiece(3, 2, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[0][2] = FieldStates.fs_Block;
        piece.Blocks[1][0] = FieldStates.fs_Block;
        return piece;
    },
    Create_J_Piece: function (color)
    {
        var piece = this.CreatePiece(3, 2, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[0][2] = FieldStates.fs_Block;
        piece.Blocks[1][2] = FieldStates.fs_Block;
        return piece;
    },
    Create_T_Piece: function (color)
    {
        var piece = this.CreatePiece(3, 2, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[0][2] = FieldStates.fs_Block;
        piece.Blocks[1][1] = FieldStates.fs_Block;
        return piece;
    },
    Create_O_Piece: function (color)
    {
        var piece = this.CreatePiece(2, 2, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[1][0] = FieldStates.fs_Block;
        piece.Blocks[1][1] = FieldStates.fs_Block;
        return piece;
    },
    Create_I_Piece: function (color)
    {
        var piece = this.CreatePiece(4, 1, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[0][2] = FieldStates.fs_Block;
        piece.Blocks[0][3] = FieldStates.fs_Block;
        return piece;
    },
    Create_S_Piece: function (color)
    {
        var piece = this.CreatePiece(3, 2, color);
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[0][2] = FieldStates.fs_Block;
        piece.Blocks[1][0] = FieldStates.fs_Block;
        piece.Blocks[1][1] = FieldStates.fs_Block;
        return piece;
    },
    Create_Z_Piece: function (color)
    {
        var piece = this.CreatePiece(3, 2, color);
        piece.Blocks[0][0] = FieldStates.fs_Block;
        piece.Blocks[0][1] = FieldStates.fs_Block;
        piece.Blocks[1][1] = FieldStates.fs_Block;
        piece.Blocks[1][2] = FieldStates.fs_Block;
        return piece;
    },
    Random: function (min, max)
    {
        var Ret = -1;
        do
        {
            Ret = min + Math.floor(Math.random() * (max + 1));
        } while (Ret === this.LastRandomNumber)

        this.LastRandomNumber = Ret;

        return Ret;
    },
    //Vygenerovat nahodnu kocku s nahodnou farbou
    Create_Random_Piece: function ()
    {
        //Typ kocky
        var pType = this.Random(0, 6);
        //Index farby
        var pColor = this.img_colors[this.Random(0, this.img_colors.length - 1)];

        switch (pType)
        {
            default:
            case 0:
                {
                    return this.Create_L_Piece(pColor);
                }
                break;
            case 1:
                {
                    return this.Create_J_Piece(pColor);
                }
                break;
            case 2:
                {
                    return this.Create_T_Piece(pColor);
                }
                break;
            case 3:
                {
                    return this.Create_O_Piece(pColor);
                }
                break;
            case 4:
                {
                    return this.Create_I_Piece(pColor);
                }
                break;
            case 5:
                {
                    return this.Create_S_Piece(pColor);
                }
                break;
            case 6:
                {
                    return this.Create_Z_Piece(pColor);
                }
                break;
        }
    },
    DrawPiece: function (piece, InPreview = false)
    {
        var Left = 0;
        var Top = 0;

        for (var i = 0; i < piece.Height; i++)
        {
            for (var j = 0; j < piece.Width; j++)
            {
                var FieldType = piece.Blocks[i][j];

                Top = piece.Top + i;
                Left = piece.Left + j;

                if (FieldType === FieldStates.fs_Block)
                {
                    if (InPreview)
                    {
                        this.DrawBlockInPreview(Top, Left, piece.Color);
                    } else
                    {
                        this.DrawBlock(Top, Left, piece.Color);
                    }
                }
            }
        }
    },
    PlacePiece: function (Piece)
    {
        var Left = Math.floor(this.GameArrayWidth / 2 - Piece.Width / 2);
        var Top = 0;
        Piece.Left = Left;
        Piece.Top = Top;

        //Overime aktualne zamyslanu poziciu
        if (this.CheckCollision(Piece, 0, 0))
        {
            this.GameOver();

        } else
        {
            this.ActivePiece = Piece;
            this.DrawPiece(Piece);
        }
    },
    NewPiece: function ()
    {
        var NewPiece = this.NextPiece;
        if (NewPiece === null)
        {
            NewPiece = this.Create_Random_Piece();
        }

        var Next = this.Create_Random_Piece();
        this.SetNextPiece(Next);

        this.PlacePiece(NewPiece);
    },
    EreasePiece: function (piece)
    {
        var xPos = 0;
        var yPos = 0;

        for (var i = 0; i < piece.Height; i++)
        {
            for (var j = 0; j < piece.Width; j++)
            {
                yPos = piece.Top + i;
                xPos = piece.Left + j;
                if (piece.Blocks[i][j] === FieldStates.fs_Block)
                {
                    this.DrawBlock(yPos, xPos, this.img_empty);
                }
            }
        }
    },
    MovePiece: function (top, left, canDock)
    {
        var piece = this.ActivePiece;
        if (this.CheckCollision(piece, top, left))
        {
            if (canDock)
            {
                this.DockPiece(piece);
            }
        } else
        {
            this.EreasePiece(piece);
            piece.Top += top;
            piece.Left += left;

            //Prekreslenie 
            this.DrawPiece(piece);
        }
    },
    GameUpdateRequest: function ()
    {
        var LevelInterval = Math.round(1000 / (this.Level + 1));

        if (LevelInterval <= 0)
            LevelInterval = 50;

        var ActualDate = new Date();
        var Interval = LevelInterval - (ActualDate - this.LastUpdate);
        if (Interval < 0)
            Interval = 0;

        window.setTimeout("Game.GameUpdate();", Interval);
    },
    GameUpdate: function ()
    {
        if (this.Status === GameStates.gs_Play)
        {
            this.LastUpdate = new Date();
            this.MoveNext();

            if (this.Status === GameStates.gs_Play)
            {
                this.GameUpdateRequest();
            }
        }
    },
    MoveNext: function ()
    {
        var Piece = this.ActivePiece;
        if (Piece.Top + Piece.Height === this.GameArrayHeight)
        { //Pristali sme na spodku hracej plochy
            this.DockPiece(Piece);
            return;
        } else //Nie sme na spodku
        {
            this.MovePiece(1, 0, true);
        }
    },
    //Ukotvit kocku na hraciu plochu
    DockPiece: function (piece)
    {
        var Top = 0;
        var Left = 0;

        //Kolko riadkov bolo vyplnenych
        var Lines = 0;

        for (var i = 0; i < piece.Height; i++)
        {
            Top = piece.Top + i;

            for (var j = 0; j < piece.Width; j++)
            {
                Left = piece.Left + j;
                if (piece.Blocks[i][j] === FieldStates.fs_Block)
                {
                    this.GameArray[Top][Left] = piece.Blocks[i][j];
                }
            }

            if (this.CheckForFullLine(Top))
            {
                Lines++;
                this.ClearLine(Top);
            }
        }

        if (Lines > 0)
        {
            //Zapisat riadky do score
            this.WriteLinesToScore(Lines);
        }

        this.NewPiece();
    },
    CheckCollision: function (piece, top, left)
    {
        var Top = 0;
        var Left = 0;

        for (var i = 0; i < piece.Height; i++)
        {
            for (var j = 0; j < piece.Width; j++)
            {
                Top = piece.Top + i + top;
                Left = piece.Left + j + left;

                if (piece.Blocks[i][j] === FieldStates.fs_Block
                        && this.GameArray[Top][Left] === piece.Blocks[i][j])
                {
                    return true;
                }
            }
        }
        return false;
    },
    RotatePiece: function (piece)
    {
        var newpiece = this.CreatePiece(piece.Height, piece.Width, piece.Color);
        var newLeft = 0;
        var widthDiff = newpiece.Width - piece.Width;

        if (widthDiff === 0)
        {
            newLeft = piece.Left;
        } else
        {
            if (newpiece.Width > piece.Width)
            {
                newLeft = piece.Left - Math.floor(newpiece.Width / 2);
            } else if (newpiece.Width < piece.Width)
            {
                newLeft = piece.Left + Math.floor(piece.Width / 2);
            }
        }

        if (newLeft < 0)
            newLeft = 0;

        if (newLeft + newpiece.Width > this.GameArrayWidth)
        {
            newLeft = this.GameArrayWidth - newpiece.Width;
        }

        newpiece.Top = piece.Top;
        newpiece.Left = newLeft;


        for (var i = 0; i < piece.Height; i++)
        {
            for (var j = 0; j < piece.Width; j++)
            {
                //V smere hodinovych ruciciek
                newpiece.Blocks[j][piece.Height - 1 - i] = piece.Blocks[i][j];
            }
        }

        //Zrotovana kocka musi vojst do priestoru
        if (this.CheckCollision(newpiece, 0, 0))
        {
            return;
        }

        //Zrotovana kocka nesmie prejst cez spodnu hranicu hracieho pola
        if (newpiece.Top + newpiece.Height > this.GameArrayHeight)
        {
            return;
        }

        //Vymaz starej kocky
        this.EreasePiece(piece);

        this.ActivePiece = newpiece;
        //Prekreslenie 
        this.DrawPiece(newpiece);
    },
    GetLowerPosition: function (piece)
    {
        var Top = piece.Top;
        var MaxTop = this.GameArrayHeight - piece.Height;

        while (Top <= MaxTop)
        {
            if (this.CheckCollision(piece, Top - piece.Top, 0))
            { //Ak doslo ku kolizii, tak posledna platna pozicia je jednu pozadu               
                return Top - 1;
            }

            Top++;
        }

        return Top - 1;
    },
    GameOver: function ()
    {
        this.Status = GameStates.gs_Stop;

        this.DrawGameField();

        this.Context.drawImage(this.img_gameover, this.GameOverLeft, this.GameOverTop);
    },
    CheckForFullLine: function (Top)
    {
        var IsFull = true;
        for (var j = 0; j < this.GameArrayWidth; j++)
        {
            if (this.GameArray[Top][j] !== FieldStates.fs_Block)
            {
                IsFull = false;
                break;
            }
        }

        return IsFull;
    },
    ClearLine: function (Top)
    {
        var RealTop = this.GetRealPosition(Top);
        var imgData = this.Context.getImageData(0, 1, this.GameAreaWidth, RealTop - this.BorderWidth);
        this.Context.putImageData(imgData, 0, 0 + this.GameFieldWidth + this.BorderWidth);

        for (var i = Top; i > 0; i--)
        {
            this.GameArray[i] = this.GameArray[i - 1];
        }

        //Novy prazdny riadok na zaciatok
        this.GameArray[0] = this.CreateEmptyLine();

        for (var j = 0; j < this.GameArrayWidth; j++)
        {
            this.GameArray[0][j] = FieldStates.fs_Empty;
            this.DrawBlock(0, j, this.img_empty);
        }
    },
    WriteLinesToScore: function (Lines)
    {
        if (Lines <= 0)
            return;

        //Aktualizacia score
        var LinesScore = 0;
        switch (Lines)
        {
            case 1:
                LinesScore = this.Line1Score;
                break;
            case 2:
                LinesScore = this.Line2Score;
                break;
            case 3:
                LinesScore = this.Line3Score;
                break;
            case 4:
                LinesScore = this.Line4Score;
                break;
            default:
            {
                //Neplatny pocet
                return;
            }
        }

        this.Score += LinesScore * (this.Level - 1) + LinesScore;
        this.LinesCounter += Lines;
        this.Level = Math.ceil(this.LinesCounter / 10);

        this.UpdateScore();

        //Prehrat zvuk vyplnenie riadkov
        this.PlaySound(this.snd_lines);

    },
    UpdateScore: function ()
    {
        this.DrawDigits(this.ScoreLeft, this.ScoreTop, this.Score);
        this.DrawDigits(this.LevelLeft, this.LevelTop, this.Level);
    },

    DrawDigits: function (Left, Top, Value)
    {
        //Textove vyjadrenie cisla
        var sValue = Value.toString();

        //Najskor vymaz stareho obsahu
        this.Context.fillStyle = this.BackColor;
        this.Context.fillRect(Left, Top, this.ContextWidth - Left, this.DigitHeight);

        //Aktualna pozicia na kreslenie
        var LeftPos = Left;

        //Pocet znakov v retazci
        var NumCount = sValue.length;

        var Img = null;
        for (var i = 0; i < NumCount; i++)
        {
            var char = sValue.charAt(i);

            if (!isNaN(char))
            {
                var ichar = parseInt(char);
                if (ichar >= 0 && ichar <= 9)
                {
                    Img = this.img_digits[ichar];
                    this.Context.drawImage(Img, LeftPos, Top);
                }
            }

            LeftPos += this.DigitWidth + this.DigitPadding;
        }
    },
    SetNextPiece: function (piece)
    {
        this.NextPiece = piece;

        var PreviewSize = 4 * this.GameFieldWidth + 3;
        //Vyplnenie celej plochy farbou pozadia
        this.Context.fillStyle = this.BackColor;
        this.Context.fillRect(this.PreviewLeft, this.PreviewTop, PreviewSize, PreviewSize);


        this.DrawPiece(this.NextPiece, true);
    },
    PlaySound: function (snd)
    {
        snd.currentTime = 0;
        snd.play();
    }
};